
#ifndef __ES_MATH_H__
#define __ES_MATH_H__

#include "es_type.h"
#include <unistd.h>

#ifdef __cplusplus
#if __cplusplus
extern "C" {
#endif
#endif /* __cplusplus */

/******************************************************************************
** ES_ABS(x)                 absolute value of x
** ES_SIGN(x)                sign of x
** ES_CMP(x,y)               0 if x==y; 1 if x>y; -1 if x<y
******************************************************************************/
#define ES_ABS(x) ((x) >= 0 ? (x) : (-(x)))
#define ES_SIGN(x) ((x) >= 0 ? 1 : -1)
#define ES_CMP(x, y) (((x) == (y)) ? 0 : (((x) > (y)) ? 1 : -1))

/******************************************************************************
** ES_MAX2(x,y)              maximum of x and y
** ES_MIN2(x,y)              minimum of x and y
** ES_MAX3(x,y,z)            maximum of x, y and z
** ES_MIN3(x,y,z)            minimun of x, y and z
** ES_MEDIAN(x,y,z)          median of x,y,z
** ES_MEAN2(x,y)             mean of x,y
******************************************************************************/
#define ES_MAX2(x, y) ((x) > (y) ? (x) : (y))
#define ES_MIN2(x, y) ((x) < (y) ? (x) : (y))
#define ES_MAX3(x, y, z) ((x) > (y) ? ES_MAX2(x, z) : ES_MAX2(y, z))
#define ES_MIN3(x, y, z) ((x) < (y) ? ES_MIN2(x, z) : ES_MIN2(y, z))
#define ES_MEDIAN(x, y, z) (((x) + (y) + (z)-ES_MAX3(x, y, z)) - ES_MIN3(x, y, z))
#define ES_MEAN2(x, y) (((x) + (y)) >> 1)

/******************************************************************************
** ES_CLIP3(x,min,max)           clip x within [min,max]
** ES_WRAP_MAX(x,max,min)        wrap to min if x equal max
** ES_WRAP_MIN(x,min,max)        wrap to max if x equal min
** ES_VALUE_BETWEEN(x,min.max)   True if x is between [min,max] inclusively.
******************************************************************************/
#define ES_CLIP_MIN(x, min) (((x) >= min) ? (x) : min)
#define ES_CLIP3(x, min, max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define ES_CLIP_MAX(x, max) ((x) > (max) ? (max) : (x))
#define ES_WRAP_MAX(x, max, min) ((x) >= (max) ? (min) : (x))
#define ES_WRAP_MIN(x, min, max) ((x) <= (min) ? (max) : (x))
#define ES_VALUE_BETWEEN(x, min, max) (((x) >= (min)) && ((x) <= (max)))

/******************************************************************************
** ES_MULTI_OF_2_POWER(x,a)  whether x is multiple of a(a must be power of 2)
** ES_ALIGN_DOWN(x,a)        floor x to multiple of a(a must be power of 2)
** ES_ALIGN_UP(x, a)         align x to multiple of a
**
** Example:
** ES_ALIGN_UP(5,4) = 8
** ES_ALIGN_DOWN(5,4) = 4
******************************************************************************/
#define ES_MULTI_OF_2_POWER(x, a) (!((x) & ((a)-1)))
#define ES_ESCEILING(x, a) (((x) + (a)-1) / (a))

#define ES_ALIGN_UP(x, a) ((((x) + ((a)-1)) / (a)) * (a))
#define ES_ALIGN_DOWN(x, a) (((x) / (a)) * (a))

#define ES_DIV_UP(x, a) (((x) + ((a)-1)) / a)

/******************************************************************************
** Get the span between two unsinged number, such as
** ES_SPAN(ES_U32, 100, 200) is 200 - 100 = 100
** ES_SPAN(ES_U32, 200, 100) is 0xFFFFFFFF - 200 + 100
** ES_SPAN(ES_U64, 200, 100) is 0xFFFFFFFFFFFFFFFF - 200 + 100
******************************************************************************/
#define ES_SPAN(type, begin, end)                              \
    ({                                                         \
        type b = (begin);                                      \
        type e = (end);                                        \
        (type)((b >= e) ? (b - e) : (b + ((~((type)0)) - e))); \
    })

/******************************************************************************
** ES_ENDIAN32(x,y)              little endian <---> big endian
** ES_IS_LITTLE_END()            whether the system is little end mode
******************************************************************************/
#define ES_ENDIAN32(x) \
    (((x) << 24) | (((x) & 0x0000ff00) << 8) | (((x) & 0x00ff0000) >> 8) | (((x) >> 24) & 0x000000ff))

/******************************************************************************
** ES_ENDIAN16(x,y)              little endian <---> big endian
** ES_IS_LITTLE_END()            whether the system is little end mode
******************************************************************************/
#define ES_ENDIAN16(x) ((((x) << 8) & 0xff00) | (((x) >> 8) & 255))

__inline static ES_BOOL ES_IS_LITTLE_END(void) {
    union unEND_TEST_U {
        ES_CHAR cTest[4];
        ES_U32 test;
    } unEndTest;

    unEndTest.cTest[0] = 0x01;
    unEndTest.cTest[1] = 0x02;
    unEndTest.cTest[2] = 0x03;
    unEndTest.cTest[3] = 0x04;

    return (unEndTest.test > 0x01020304) ? (ES_TRUE) : (ES_FALSE);
}

/******************************************************************************
** ES_FRACTION32(de,nu)          fraction: nu(minator) / de(nominator).
** ES_NUMERATOR32(x)             of x(x is fraction)
** ES_DENOMINATOR32(x)           Denominator of x(x is fraction)

** represent fraction in 32 bit. LSB 16 is numerator, MSB 16 is denominator
** It is integer if denominator is 0.
******************************************************************************/
#define ES_FRACTION32(de, nu) (((de) << 16) | (nu))
#define ES_NUMERATOR32(x) ((x) & 0xffff)
#define ES_DENOMINATOR32(x) ((x) >> 16)

/******************************************************************************
** ES_RGB(r,g,b)    assemble the r,g,b to 24bit color
** ES_RGB_R(c)      get RED   from 24bit color
** ES_RGB_G(c)      get GREEN from 24bit color
** ES_RGB_B(c)      get BLUE  from 24bit color
******************************************************************************/
#define ES_RGB(r, g, b) ((((r) & 0xff) << 16) | (((g) & 0xff) << 8) | ((b) & 0xff))
#define ES_RGB_R(c) (((c) & 0xff0000) >> 16)
#define ES_RGB_G(c) (((c) & 0xff00) >> 8)
#define ES_RGB_B(c) ((c) & 0xff)

/******************************************************************************
** ES_YUV(y,u,v)    assemble the y,u,v to 30bit color
** ES_YUV_Y(c)      get Y from 30bit color
** ES_YUV_U(c)      get U from 30bit color
** ES_YUV_V(c)      get V from 30bit color
******************************************************************************/
#define ES_YUV(y, u, v) ((((y) & 0x03ff) << 20) | (((u) & 0x03ff) << 10) | ((v) & 0x03ff))
#define ES_YUV_Y(c) (((c) & 0x3ff00000) >> 20)
#define ES_YUV_U(c) (((c) & 0x000ffc00) >> 10)
#define ES_YUV_V(c) ((c) & 0x000003ff)

/******************************************************************************
** ES_YUV_8BIT(y,u,v)    assemble the y,u,v to 24bit color
** ES_YUV_8BIT_Y(c)      get Y from 24bit color
** ES_YUV_8BIT_U(c)      get U from 24bit color
** ES_YUV_8BIT_V(c)      get V from 24bit color
******************************************************************************/
#define ES_YUV_8BIT(y, u, v) ((((y) & 0xff) << 16) | (((u) & 0xff) << 8) | ((v) & 0xff))
#define ES_YUV_8BIT_Y(c) (((c) & 0xff0000) >> 16)
#define ES_YUV_8BIT_U(c) (((c) & 0xff00) >> 8)
#define ES_YUV_8BIT_V(c) ((c) & 0xff)

/******************************************************************************
** ES_Rgb2Yc(r, g, b, *y, *u, *u)    convert r,g,b to y,u,v
** ES_Rgb2Yuv(rgb)                   convert rgb to yuv
******************************************************************************/
__inline static ES_VOID ES_Rgb2Yc(ES_U16 r, ES_U16 g, ES_U16 b, ES_U16 *py, ES_U16 *pcb, ES_U16 *pcr) {
    /* Y */
    *py = (ES_U16)((((r * 66 + g * 129 + b * 25) >> 8) + 16) << 2);

    /* Cb */
    *pcb = (ES_U16)(((((b * 112 - r * 38) - g * 74) >> 8) + 128) << 2);

    /* Cr */
    *pcr = (ES_U16)(((((r * 112 - g * 94) - b * 18) >> 8) + 128) << 2);
}

__inline static ES_U32 ES_Rgb2Yuv(ES_U32 rgb) {
    ES_U16 y, u, v;

    ES_Rgb2Yc(ES_RGB_R(rgb), ES_RGB_G(rgb), ES_RGB_B(rgb), &y, &u, &v);

    return ES_YUV(y, u, v);
}

__inline static ES_VOID ES_Rgb2Yc_full(ES_U16 r, ES_U16 g, ES_U16 b, ES_U16 *py, ES_U16 *pcb, ES_U16 *pcr) {
    ES_U16 py_temp, pcb_temp, pcr_temp;

    py_temp = (ES_U16)(((r * 76 + g * 150 + b * 29) >> 8) * 4);
    pcb_temp = (ES_U16)(ES_CLIP_MIN(((((b * 130 - r * 44) - g * 86) >> 8) + 128), 0) * 4);
    pcr_temp = (ES_U16)(ES_CLIP_MIN(((((r * 130 - g * 109) - b * 21) >> 8) + 128), 0) * 4);

    *py = ES_MAX2(ES_MIN2(py_temp, 1023), 0);
    *pcb = ES_MAX2(ES_MIN2(pcb_temp, 1023), 0);
    *pcr = ES_MAX2(ES_MIN2(pcr_temp, 1023), 0);
}

__inline static ES_U32 ES_Rgb2Yuv_full(ES_U32 rgb) {
    ES_U16 y, u, v;

    ES_Rgb2Yc_full(ES_RGB_R(rgb), ES_RGB_G(rgb), ES_RGB_B(rgb), &y, &u, &v);

    return ES_YUV(y, u, v);
}

/******************************************************************************
** ES_Rgb2Yc_8BIT(r, g, b, *y, *u, *u)    convert r,g,b to y,u,v
** ES_Rgb2Yuv_8BIT(rgb)                   convert rgb to yuv
******************************************************************************/
__inline static ES_VOID ES_Rgb2Yc_8BIT(ES_U8 r, ES_U8 g, ES_U8 b, ES_U8 *py, ES_U8 *pcb, ES_U8 *pcr) {
    /* Y */
    *py = (ES_U8)(((r * 66 + g * 129 + b * 25) >> 8) + 16);

    /* Cb */
    *pcb = (ES_U8)((((b * 112 - r * 38) - g * 74) >> 8) + 128);

    /* Cr */
    *pcr = (ES_U8)((((r * 112 - g * 94) - b * 18) >> 8) + 128);
}

__inline static ES_U32 ES_Rgb2Yuv_8BIT(ES_U32 rgb) {
    ES_U8 y, u, v;

    ES_Rgb2Yc_8BIT(ES_RGB_R(rgb), ES_RGB_G(rgb), ES_RGB_B(rgb), &y, &u, &v);

    return ES_YUV_8BIT(y, u, v);
}

__inline static ES_VOID ES_Rgb2Yc_full_8BIT(ES_U8 r, ES_U8 g, ES_U8 b, ES_U8 *py, ES_U8 *pcb, ES_U8 *pcr) {
    ES_S16 py_temp, pcb_temp, pcr_temp;

    py_temp = (r * 76 + g * 150 + b * 29) >> 8;
    pcb_temp = (((b * 130 - r * 44) - g * 86) >> 8) + 128;
    pcr_temp = (((r * 130 - g * 109) - b * 21) >> 8) + 128;

    *py = ES_MAX2(ES_MIN2(py_temp, 255), 0);
    *pcb = ES_MAX2(ES_MIN2(pcb_temp, 255), 0);
    *pcr = ES_MAX2(ES_MIN2(pcr_temp, 255), 0);
}

__inline static ES_U32 ES_Rgb2Yuv_full_8BIT(ES_U32 rgb) {
    ES_U8 y, u, v;

    ES_Rgb2Yc_full_8BIT(ES_RGB_R(rgb), ES_RGB_G(rgb), ES_RGB_B(rgb), &y, &u, &v);

    return ES_YUV_8BIT(y, u, v);
}

/*******************************************************************************
** ES_FpsControl Useing Sample:
**  FPS_CTRL_S g_fpsCtrl;
**
**  Take 12 frame uniform in 25.
**  ES_InitFps(&g_fpsCtrl, 25, 12);
**
**  {
**       if(ES_FpsControl(&g_fpsCtrl)) printf("Yes, this frmae should be token");
**  }
**
*******************************************************************************/
typedef struct esFPS_CTRL_S {
    ES_U32 ffps;   /* Full frame rate    */
    ES_U32 tfps;   /* Target frame rate  */
    ES_U32 frmKey; /* update key frame   */
} FPS_CTRL_S;

__inline static ES_VOID ES_InitFps(FPS_CTRL_S *pFrmCtrl, ES_U32 fullFps, ES_U32 tagFps) {
    pFrmCtrl->ffps = fullFps;
    pFrmCtrl->tfps = tagFps;
    pFrmCtrl->frmKey = 0;
}

__inline static ES_BOOL ES_FpsControl(FPS_CTRL_S *pFrmCtrl) {
    ES_BOOL bReturn = ES_FALSE;

    pFrmCtrl->frmKey += pFrmCtrl->tfps;
    if (pFrmCtrl->frmKey >= pFrmCtrl->ffps) {
        pFrmCtrl->frmKey -= pFrmCtrl->ffps;
        bReturn = ES_TRUE;
    }

    return bReturn;
}

__inline static ES_U32 ES_GetLowAddr(ES_U64 phyaddr) {
    return (ES_U32)phyaddr;
}

__inline static ES_U32 ES_GetHighAddr(ES_U64 phyaddr) {
    return (ES_U32)(phyaddr >> 32);
}

#define es_usleep(usec) \
    do {                \
        usleep(usec);   \
    } while (0)

#ifdef __cplusplus
#if __cplusplus
}
#endif
#endif /* __cplusplus */

#endif /* __ES_MATH_H__ */
